# -*- coding: utf-8 -*-
"""
Created on Fri Jul  2 10:10:55 2021

@author: pearsonra
"""

import urllib
import pathlib
import requests
import boto3
import botocore
import botocore.client
import typing
import geopandas
import abc
from . import geometry


class S3QueryBase(abc.ABC):
    """A class to manage fetching LiDAR data from Open Topography.

    API details for querying datasets within a search rectangle in a AWS style bucket

    All datasets within a search polygon may be downloaded; or datasets may be selected
    by name (either within a search  polygon or the entire dataset).
    """

    @property
    @abc.abstractmethod
    def NETLOC_DATA():
        """This should be instantiated in the base class. Provide the netloc of the data
        service."""

        raise NotImplementedError("NETLOC_API must be instantiated in the child class")

    @property
    @abc.abstractmethod
    def OT_BUCKET():
        """This should be instantiated in the base class. Provide the netloc of the data
        service."""

        raise NotImplementedError("NETLOC_API must be instantiated in the child class")

    SCHEME = "https"
    NETLOC_API = "portal.opentopography.org"
    PATH_API = "/API/otCatalog"
    OT_CRS = "EPSG:4326"

    def __init__(
        self,
        cache_path: typing.Union[str, pathlib.Path],
        search_polygon: geopandas.geodataframe.GeoDataFrame = None,
        redownload_files: bool = False,
        download_limit_gbytes: typing.Union[int, float] = 100,
        verbose: bool = False,
    ):
        """Define the cache_path (or location where the data will be downloaded). Other
        attributes are optional. If a
        search_polygon is not specified then datasets must be downloaded by name."""

        self.search_polygon = search_polygon
        self.cache_path = pathlib.Path(cache_path)
        self.redownload_files_bool = redownload_files
        self.download_limit_gbytes = download_limit_gbytes
        self.verbose = verbose

        self._dataset_prefixes = None

    def _to_gbytes(self, bytes_number: int):
        """convert bytes into gigabytes"""

        return bytes_number / 1024 / 1024 / 1024

    def run(self, dataset_name: str = None):
        """Download LiDAR dataset(s) either within a search_polygon, by name, or both"""

        if dataset_name is not None:
            self.download_dataset_by_name(dataset_name)
        elif self.search_polygon is not None:
            self.download_datasets_in_polygon()
        else:
            print(
                "Both the search_polygon and dataset_name are None. Either a "
                "dataset_name of search polygon needs to be specified if any datasets "
                "are to be downloaded from OpenTopography. Please specify a "
                "search_polygon during construction, or a dataset_name during run."
            )

    def download_datasets_in_polygon(self):
        """Download all LiDAR datasets within the search polygon"""

        ot_endpoint_url = urllib.parse.urlunparse(
            (self.SCHEME, self.NETLOC_DATA, "", "", "", "")
        )
        client = boto3.client(
            "s3",
            endpoint_url=ot_endpoint_url,
            config=botocore.config.Config(signature_version=botocore.UNSIGNED),
        )

        self._dataset_prefixes = []
        json_response = self.query_for_datasets_inside_catchment()

        # cycle through each dataset within a region
        for json_dataset in json_response["Datasets"]:
            dataset_prefix = json_dataset["Dataset"]["alternateName"]
            self._dataset_prefixes.append(dataset_prefix)

            self.download_dataset(dataset_prefix, client)

    def download_dataset_by_name(self, dataset_name: str):
        """Download a LiDAR dataset by name after checking it is within any specified
        search polygon"""

        ot_endpoint_url = urllib.parse.urlunparse(
            (self.SCHEME, self.NETLOC_DATA, "", "", "", "")
        )
        client = boto3.client(
            "s3",
            endpoint_url=ot_endpoint_url,
            config=botocore.config.Config(signature_version=botocore.UNSIGNED),
        )

        self._dataset_prefixes = []
        if self.search_polygon is None:
            # No search polygon so download the specified dataset
            self._dataset_prefixes.append(dataset_name)
            self.download_dataset(dataset_name, client)
        else:
            # Only download if the specified dataset name is in the search polygon
            json_response = self.query_for_datasets_inside_catchment()
            for json_dataset in json_response["Datasets"]:
                dataset_prefix = json_dataset["Dataset"]["alternateName"]
                if dataset_prefix == dataset_name:
                    self._dataset_prefixes.append(dataset_name)
                    self.download_dataset(dataset_name, client)
                    break

    def query_for_datasets_inside_catchment(self):
        """Function to check for data in search region using the otCatalogue API
        https://portal.opentopography.org/apidocs/#/Public/getOtCatalog"""

        catchment_bounds = self.search_polygon.geometry.to_crs(self.OT_CRS).bounds
        api_query = {
            "productFormat": "PointCloud",
            "minx": catchment_bounds["minx"].min(),
            "miny": catchment_bounds["miny"].min(),
            "maxx": catchment_bounds["maxx"].max(),
            "maxy": catchment_bounds["maxy"].max(),
            "detail": False,
            "outputFormat": "json",
            "inlcude_federated": True,
        }

        data_url = urllib.parse.urlunparse(
            (self.SCHEME, self.NETLOC_API, self.PATH_API, "", "", "")
        )

        response = requests.get(data_url, params=api_query, stream=True)
        response.raise_for_status()
        return response.json()

    def download_dataset(self, dataset_prefix, client):
        """Download all files within an optional search polygon of a given
        dataset_prefix"""

        if self.verbose:
            print(f"Check files in dataset {dataset_prefix}")
        tile_info = self._get_dataset_tile_names(client, dataset_prefix)

        # check download size limit is not exceeded
        lidar_size_bytes = self._calculate_dataset_download_size(
            client, dataset_prefix, tile_info
        )

        assert self._to_gbytes(lidar_size_bytes) < self.download_limit_gbytes, (
            "The size of the LiDAR to be downloaded is "
            f"{self._to_gbytes(lidar_size_bytes)}GB, which greater than the specified "
            f"download limit of {self.download_limit_gbytes}GB. Please free up some and"
            " try again."
        )

        # check for tiles and download as needed
        self._download_tiles_in_catchment(client, dataset_prefix, tile_info)

    def _get_dataset_tile_names(self, client, dataset_prefix):
        """Check for the tile index shapefile and download as needed, then load in and
        trim to the catchment to determine which data tiles to download."""

        file_prefix = f"{dataset_prefix}/{dataset_prefix}_TileIndex.zip"
        local_file_path = self.cache_path / file_prefix

        # Download the file if needed
        if self.redownload_files_bool or not local_file_path.exists():

            # ensure folder exists before download
            local_file_path.parent.mkdir(parents=True, exist_ok=True)
            try:
                client.download_file(self.OT_BUCKET, file_prefix, str(local_file_path))
            except botocore.exceptions.ClientError as e:
                f"An error occured during download {file_prefix}, The error is {e}"
        # load in tile information
        tile_info = geometry.TileInfo(local_file_path, self.search_polygon)

        return tile_info

    def _calculate_dataset_download_size(self, client, dataset_prefix, tile_info):
        """Sum up the size of the LiDAR data in catchment"""
        lidar_size_bytes = 0

        for tile_url in tile_info.urls:
            # drop the OT_BUCKET from the URL path to get the file_prefix
            file_prefix = pathlib.Path(
                *pathlib.Path(urllib.parse.urlparse(tile_url).path).parts[2:]
            )
            local_path = self.cache_path / file_prefix
            if self.redownload_files_bool or not local_path.exists():

                try:
                    response = client.head_object(
                        Bucket=self.OT_BUCKET, Key=str(file_prefix.as_posix())
                    )
                    lidar_size_bytes += response["ContentLength"]

                    if self.verbose:
                        print(
                            f"checking size: {file_prefix}: {response['ContentLength']}"
                            f", total (GB): {self._to_gbytes(lidar_size_bytes)}"
                        )
                except botocore.exceptions.ClientError as e:
                    f"An error occured during access of {file_prefix}, The error is {e}"
        return lidar_size_bytes

    def _download_tiles_in_catchment(self, client, dataset_prefix, tile_info):
        """Download the LiDAR data within the catchment"""

        for url in tile_info.urls:
            # drop the OT_BUCKET from the URL path to get the file_prefix
            file_prefix = pathlib.Path(
                *pathlib.Path(urllib.parse.urlparse(url).path).parts[2:]
            )
            local_path = self.cache_path / file_prefix

            # ensure folder exists before download - in case its in a subdirectory that
            # hasn't been created yet
            local_path.parent.mkdir(parents=True, exist_ok=True)

            if self.redownload_files_bool or not local_path.exists():
                if self.verbose:
                    print(f"Downloading file: {file_prefix}")
                try:
                    client.download_file(
                        self.OT_BUCKET, str(file_prefix.as_posix()), str(local_path)
                    )
                except botocore.exceptions.ClientError as e:
                    f"An error occured during download {file_prefix}, The error is {e}"

    @property
    def dataset_prefixes(self):
        """Get the dataset names of all datasets downloaded by this object."""

        assert self._dataset_prefixes is not None, (
            "The run command needs to be called before 'dataset_prefixes' can "
            + "be called."
        )

        return self._dataset_prefixes


class OpenTopography(S3QueryBase):
    """A class to manage fetching LiDAR data from Open Topography

    API details for querying datasets within a search rectangle at:
        https://portal.opentopography.org/apidocs/#/Public/getOtCatalog
    Information for making a `bulk download` of a dataset using the AWS S3 protocol can
    be found by clicking on bulk download under any public dataset.

    All datasets within a search polygon may be downloaded; or datasets may be selected
    by name (either within a search polygon or the entire dataset).
    """

    NETLOC_DATA = "opentopography.s3.sdsc.edu"
    OT_BUCKET = "pc-bulk"



# -*- coding: utf-8 -*-
"""
Created on Fri Nov  7 10:10:55 2022

@author: pearsonra
"""

import urllib
import requests
import numpy
import geopandas
import abc
import typing
import pathlib
import logging
import time
import io
import zipfile


class KoordinatesExportsQueryBase(abc.ABC):
    """An abstract class to manage fetching Raster data using the Koordinates exports
    API. Downloads the GeoTiff specified in the run routine.

    API details at: https://help.koordinates.com/site-admin-apis/export-api/

    Parameters
    ----------

    key: str
        The API key.  Must have Search, view and download exported data
        permissions.

    cache_path: pathlib.Path
        The location to download all GeoTiffs queried in the run method.

    crs: int
        The CRS EPSG code for the GeoTifss to be downloaded as.

    bounding_polygon: geopandas.geodataframe.GeoDataFrame
        An option geometry to clip the downloaded GeoTiffs within.

    """

    @property
    @abc.abstractmethod
    def NETLOC_API():
        """This should be instantiated in the base class. Provide the netloc of the data
        service."""

        raise NotImplementedError("NETLOC_API must be instantiated in the child class")

    SCHEME = "https"
    PATH = "services/api/v1"
    PATH_API_END = "/exports"
    K_CRS = "EPSG:4326"

    def __init__(
        self,
        key: str,
        cache_path: typing.Union[str, pathlib.Path],
        crs: int = None,
        bounding_polygon: geopandas.geodataframe.GeoDataFrame = None,
    ):
        """Load in the wfs key and CRS/bounding_polygon if specified. Specify the layer
        to import during run."""

        self.key = key
        self.cache_path = pathlib.Path(cache_path)
        self.bounding_polygon = bounding_polygon
        self.crs = crs

        self.base_url = urllib.parse.urlunparse(
            (
                self.SCHEME,
                self.NETLOC_API,
                self.PATH,
                "",
                "",
                "",
            )
        )

        self._set_up()

    def _set_up(self):
        """Ensure the bouding_polygon and CRS are in agreement."""

        # Set the crs from the bounding_polygon if it's not been set
        if self.crs is None and self.bounding_polygon is not None:
            logging.info("The download CRS is being set from the bounding_polygon")
            self.crs = self.bounding_polygon.crs.to_epsg()
        # Set the bounding_polygon crs from the crs if they differ
        if (
            self.bounding_polygon is not None
            and self.crs != self.bounding_polygon.crs.to_epsg()
        ):
            logging.info(
                "The bounding_polygon is being transformed to the specified "
                "download CRS"
            )
            self.bounding_polygon.to_crs(self.crs)
        # Enforce the bounding_polygon must be a single geometry if it exists
        if self.bounding_polygon is not None:
            assert (
                len(self.bounding_polygon) == 1
            ), "The bounding polygon must be a single geometry"
            assert (
                len(numpy.array(self.bounding_polygon.exterior.loc[0].coords)) < 1000
            ), "The bounding polygon must be lass than 1000 points"

    def run(self, layer: int) -> pathlib.Path:
        """Query for a specified layer and return a geopandas.GeoDataFrame of the vector
        features. If a polygon_boundary is specified, only return vectors passing
        through this polygon."""

        headers = {"Authorization": f"key {self.key}"}

        # Create the initial request
        api_query = {
            "crs": f"EPSG:{self.crs}",
            "formats": {"grid": "image/tiff;subtype=geotiff"},
            "items": [{"item": f"{self.base_url}/layers/{layer}/"}],
        }
        if self.bounding_polygon is not None:
            exterior = self.bounding_polygon.to_crs(self.K_CRS).exterior.loc[0]
            api_query["extent"] = {
                "type": self.bounding_polygon.type.loc[0],
                "coordinates": [list(exterior.coords)],
            }
        logging.info("Send initial request to download image")
        response = requests.post(
            url=f"{self.base_url}/exports/", headers=headers, json=api_query
        )
        query_id = response.json()["id"]

        # Check the state of your exports until the triggered raster exports completes
        logging.info("Check status of download request")
        while True:
            response = requests.get(
                f"{self.base_url}/exports/",
                headers=headers,
            )
            # find the triggered export
            element = [
                element for element in response.json() if element["id"] == query_id
            ][0]
            logging.info(f"/texport state is {element['state']}")
            if element["state"] == "processing":
                logging.info("Not complete - check again in 20s")
                time.sleep(20)
                continue
            elif element["state"] == "complete":
                logging.info("/tCompleted - move to download")
                break
            else:
                logging.warning(
                    f"Could not download raster. Ended with status {element['state']}"
                )
                return
        # Download the completed export
        logging.info(f"Downloading {element['download_url']} to {self.cache_path}")
        with requests.get(
            element["download_url"],
            headers={"Authorization": f"key {self.key}"},
            stream=True,
        ) as response:
            response.raise_for_status()
            zip_object = zipfile.ZipFile(io.BytesIO(response.content))
            zip_object.extractall(self.cache_path / f"{layer}")
        # Return the file names of the downloaded rasters
        rasters = []
        for file_name in (self.cache_path / f"{layer}").iterdir():
            if file_name.suffix == ".tif":
                rasters.append(file_name)
        return rasters


class Linz(KoordinatesExportsQueryBase):
    """A class to manage fetching Vector data from LINZ.

    LIRS data service can be accessed at: https://https://data.linz.govt.nz/

    Note that only rasters supporting the grid image/tiff geotiff are supported
    """

    NETLOC_API = "data.linz.govt.nz"


class Lris(KoordinatesExportsQueryBase):
    """A class to manage fetching Vector data from LRIS.

    LIRS data service can be accessed at: https://lris.scinfo.org.nz/

    Note that only rasters supporting the grid image/tiff geotiff are supported
    """

    NETLOC_API = "lris.scinfo.org.nz"


class StatsNz(KoordinatesExportsQueryBase):
    """A class to manage fetching Vector data from the Stats NZ datafinder.

    Stats NZ data service can be accessed at: datafinder.stats.govt.nz

    Note that only rasters supporting the grid image/tiff geotiff are supported
    """

    NETLOC_API = "datafinder.stats.govt.nz"


class KoordinatesQuery(KoordinatesExportsQueryBase):
    """A class to manage fetching Vector data from any generic data portal supporting
    WFS.

    Note that the 'geometry_name' used when making a WFS 'cql_filter' queries can vary
    between layers. You will need to specify the 'geometry_name' of the layers you want
    to download.
    """

    def __init__(
        self,
        key: str,
        netloc_url: str,
        crs: int = None,
        bounding_polygon: geopandas.geodataframe.GeoDataFrame = None,
    ):
        """Set NETLOC_API and instantiate the KoordinatesExportsQueryBase"""

        self.netloc_url = netloc_url

        # Setup the WfsQueryBase class
        super(KoordinatesQuery, self).__init__(
            key=key, crs=crs, bounding_polygon=bounding_polygon
        )

    @property
    def NETLOC_API(self):
        """Instantiate the entered netloc of the data service."""

        return self.netloc_url
